/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.gui;

import jalview.api.FeatureColourI;
import jalview.datamodel.SearchResults;
import jalview.datamodel.SearchResultsI;
import jalview.datamodel.SequenceFeature;
import jalview.datamodel.SequenceI;
import jalview.io.FeaturesFile;
import jalview.schemes.FeatureColour;
import jalview.util.ColorUtils;
import jalview.util.MessageManager;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

import javax.swing.JColorChooser;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingConstants;

/**
 * DOCUMENT ME!
 * 
 * @author $author$
 * @version $Revision$
 */
public class FeatureRenderer extends
        jalview.renderer.seqfeatures.FeatureRenderer
{
  Color resBoxColour;

  AlignmentPanel ap;

  /**
   * Creates a new FeatureRenderer object
   * 
   * @param alignPanel
   */
  public FeatureRenderer(AlignmentPanel alignPanel)
  {
    super(alignPanel.av);
    this.ap = alignPanel;
    if (alignPanel.getSeqPanel() != null
            && alignPanel.getSeqPanel().seqCanvas != null
            && alignPanel.getSeqPanel().seqCanvas.fr != null)
    {
      transferSettings(alignPanel.getSeqPanel().seqCanvas.fr);
    }
  }

  // // /////////////
  // // Feature Editing Dialog
  // // Will be refactored in next release.

  static String lastFeatureAdded;

  static String lastFeatureGroupAdded;

  static String lastDescriptionAdded;

  FeatureColourI oldcol, fcol;

  int featureIndex = 0;

  /**
   * Presents a dialog allowing the user to add new features, or amend or delete
   * existing features. Currently this can be on
   * <ul>
   * <li>double-click on a sequence - Amend/Delete features at position</li>
   * <li>Create sequence feature from pop-up menu on selected region</li>
   * <li>Create features for pattern matches from Find</li>
   * </ul>
   * 
   * @param sequences
   *          the sequences features are to be created on (if creating
   *          features), or a single sequence (if amending features)
   * @param features
   *          the current features at the position (if amending), or template
   *          new features with start/end position set (if creating)
   * @param create
   *          true to create features, false to amend or delete
   * @param featureType
   *          the feature type to set on new features; if null, defaults to the
   *          type of the last new feature created if any, failing that to
   *          "feature_1"
   * @param alignPanel
   * @return
   */
  protected boolean amendFeatures(final List<SequenceI> sequences,
          final List<SequenceFeature> features, boolean create,
          final AlignmentPanel alignPanel, String featureType)
  {

    featureIndex = 0;

    final JPanel mainPanel = new JPanel(new BorderLayout());
    final JTextField name = new JTextField(25);
    final JTextField source = new JTextField(25);
    final JTextArea description = new JTextArea(3, 25);
    final JSpinner start = new JSpinner();
    final JSpinner end = new JSpinner();
    start.setPreferredSize(new Dimension(80, 20));
    end.setPreferredSize(new Dimension(80, 20));
    final FeatureRenderer me = this;
    final JLabel colour = new JLabel();
    colour.setOpaque(true);
    // colour.setBorder(BorderFactory.createEtchedBorder());
    colour.setMaximumSize(new Dimension(30, 16));
    colour.addMouseListener(new MouseAdapter()
    {
      FeatureColourChooser fcc = null;

      @Override
      public void mousePressed(MouseEvent evt)
      {
        if (fcol.isSimpleColour())
        {
          Color col = JColorChooser.showDialog(Desktop.desktop,
                  MessageManager.getString("label.select_feature_colour"),
                  fcol.getColour());
          if (col != null)
          {
            fcol = new FeatureColour(col);
            updateColourButton(mainPanel, colour, new FeatureColour(col));
          }
        }
        else
        {
          if (fcc == null)
          {
            final String ft = features.get(featureIndex).getType();
            final String type = ft == null ? lastFeatureAdded : ft;
            fcc = new FeatureColourChooser(me, type);
            fcc.setRequestFocusEnabled(true);
            fcc.requestFocus();

            fcc.addActionListener(new ActionListener()
            {

              @Override
              public void actionPerformed(ActionEvent e)
              {
                fcol = fcc.getLastColour();
                fcc = null;
                setColour(type, fcol);
                updateColourButton(mainPanel, colour, fcol);
              }
            });

          }
        }
      }
    });
    JPanel gridPanel = new JPanel(new GridLayout(3, 1));

    if (!create && features.size() > 1)
    {
      /*
       * more than one feature at selected position - add a drop-down
       * to choose the feature to amend
       */
      gridPanel = new JPanel(new GridLayout(4, 1));
      JPanel choosePanel = new JPanel();
      choosePanel.add(new JLabel(MessageManager
              .getString("label.select_feature")
              + ":"));
      final JComboBox<String> overlaps = new JComboBox<String>();
      for (SequenceFeature sf : features)
      {
        String text = sf.getType() + "/" + sf.getBegin() + "-"
                + sf.getEnd() + " (" + sf.getFeatureGroup() + ")";
        overlaps.addItem(text);
      }
      choosePanel.add(overlaps);

      overlaps.addItemListener(new ItemListener()
      {
        @Override
        public void itemStateChanged(ItemEvent e)
        {
          int index = overlaps.getSelectedIndex();
          if (index != -1)
          {
            featureIndex = index;
            SequenceFeature sf = features.get(index);
            name.setText(sf.getType());
            description.setText(sf.getDescription());
            source.setText(sf.getFeatureGroup());
            start.setValue(new Integer(sf.getBegin()));
            end.setValue(new Integer(sf.getEnd()));

            SearchResultsI highlight = new SearchResults();
            highlight.addResult(sequences.get(0), sf.getBegin(),
                    sf.getEnd());

            alignPanel.getSeqPanel().seqCanvas.highlightSearchResults(highlight);

          }
          FeatureColourI col = getFeatureStyle(name.getText());
          if (col == null)
          {
            col = new FeatureColour(ColorUtils
                    .createColourFromName(name.getText()));
          }
          oldcol = fcol = col;
          updateColourButton(mainPanel, colour, col);
        }
      });

      gridPanel.add(choosePanel);
    }
    // ////////
    // ////////////////////////////////////

    JPanel namePanel = new JPanel();
    gridPanel.add(namePanel);
    namePanel.add(new JLabel(MessageManager.getString("label.name:"),
            JLabel.RIGHT));
    namePanel.add(name);

    JPanel groupPanel = new JPanel();
    gridPanel.add(groupPanel);
    groupPanel.add(new JLabel(MessageManager.getString("label.group:"),
            JLabel.RIGHT));
    groupPanel.add(source);

    JPanel colourPanel = new JPanel();
    gridPanel.add(colourPanel);
    colourPanel.add(new JLabel(MessageManager.getString("label.colour"),
            JLabel.RIGHT));
    colourPanel.add(colour);
    colour.setPreferredSize(new Dimension(150, 15));
    colour.setFont(new java.awt.Font("Verdana", Font.PLAIN, 9));
    colour.setForeground(Color.black);
    colour.setHorizontalAlignment(SwingConstants.CENTER);
    colour.setVerticalAlignment(SwingConstants.CENTER);
    colour.setHorizontalTextPosition(SwingConstants.CENTER);
    colour.setVerticalTextPosition(SwingConstants.CENTER);
    mainPanel.add(gridPanel, BorderLayout.NORTH);

    JPanel descriptionPanel = new JPanel();
    descriptionPanel.add(new JLabel(MessageManager
            .getString("label.description:"),
            JLabel.RIGHT));
    description.setFont(JvSwingUtils.getTextAreaFont());
    description.setLineWrap(true);
    descriptionPanel.add(new JScrollPane(description));

    if (!create)
    {
      mainPanel.add(descriptionPanel, BorderLayout.SOUTH);

      JPanel startEndPanel = new JPanel();
      startEndPanel.add(new JLabel(MessageManager.getString("label.start"),
              JLabel.RIGHT));
      startEndPanel.add(start);
      startEndPanel.add(new JLabel(MessageManager.getString("label.end"),
              JLabel.RIGHT));
      startEndPanel.add(end);
      mainPanel.add(startEndPanel, BorderLayout.CENTER);
    }
    else
    {
      mainPanel.add(descriptionPanel, BorderLayout.CENTER);
    }

    SequenceFeature firstFeature = features.get(0);
    if (featureType != null)
    {
      lastFeatureAdded = featureType;
    }
    else
    {
      if (lastFeatureAdded == null)
      {
        if (firstFeature.type != null)
        {
          lastFeatureAdded = firstFeature.type;
        }
        else
        {
          lastFeatureAdded = "feature_1";
        }
      }
    }

    if (lastFeatureGroupAdded == null)
    {
      if (firstFeature.featureGroup != null)
      {
        lastFeatureGroupAdded = firstFeature.featureGroup;
      }
      else
      {
        lastFeatureGroupAdded = "Jalview";
      }
    }

    if (create)
    {
      name.setText(lastFeatureAdded);
      source.setText(lastFeatureGroupAdded);
    }
    else
    {
      name.setText(firstFeature.getType());
      source.setText(firstFeature.getFeatureGroup());
    }

    start.setValue(new Integer(firstFeature.getBegin()));
    end.setValue(new Integer(firstFeature.getEnd()));
    description.setText(firstFeature.getDescription());
    updateColourButton(mainPanel, colour,
            (oldcol = fcol = getFeatureStyle(name.getText())));
    Object[] options;
    if (!create)
    {
      options = new Object[] { MessageManager.getString("label.amend"),
          MessageManager.getString("action.delete"),
          MessageManager.getString("action.cancel") };
    }
    else
    {
      options = new Object[] { MessageManager.getString("action.ok"),
          MessageManager.getString("action.cancel") };
    }

    String title = create ? MessageManager
            .getString("label.create_new_sequence_features")
            : MessageManager.formatMessage("label.amend_delete_features",
                    new String[] { sequences.get(0).getName() });

    /*
     * show the dialog
     */
    int reply = JvOptionPane.showInternalOptionDialog(Desktop.desktop,
            mainPanel, title, JvOptionPane.YES_NO_CANCEL_OPTION,
            JvOptionPane.QUESTION_MESSAGE, null, options,
            MessageManager.getString("action.ok"));

    FeaturesFile ffile = new FeaturesFile();

    if (reply == JvOptionPane.OK_OPTION && name.getText().length() > 0)
    {
      lastFeatureAdded = name.getText().trim();
      lastFeatureGroupAdded = source.getText().trim();
      lastDescriptionAdded = description.getText().replaceAll("\n", " ");
      // TODO: determine if the null feature group is valid
      if (lastFeatureGroupAdded.length() < 1)
      {
        lastFeatureGroupAdded = null;
      }
    }

    if (!create)
    {
      SequenceFeature sf = features.get(featureIndex);

      if (reply == JvOptionPane.NO_OPTION)
      {
        /*
         * NO_OPTION corresponds to the Delete button
         */
        sequences.get(0).getDatasetSequence().deleteFeature(sf);
      }
      else if (reply == JvOptionPane.YES_OPTION)
      {
        /*
         * YES_OPTION corresponds to the Amend button
         */
        boolean typeChanged = !lastFeatureAdded.equals(sf.type);
        String newType = lastFeatureAdded;
        String newFeatureGroup = lastFeatureGroupAdded;
        String newDescription = lastDescriptionAdded;

        setColour(newType, fcol);
        getFeaturesDisplayed().setVisible(newType);
        int newBegin = sf.begin;
        int newEnd = sf.end;
        try
        {
          newBegin = ((Integer) start.getValue()).intValue();
          newEnd = ((Integer) end.getValue()).intValue();
        } catch (NumberFormatException ex)
        {
          // JSpinner doesn't accept invalid format data :-)
        }

        /*
         * replace the feature by deleting it and adding a new one
         * (to ensure integrity of SequenceFeatures data store)
         */
        sequences.get(0).deleteFeature(sf);
        SequenceFeature newSf = new SequenceFeature(newType,
                newDescription, newBegin, newEnd, sf.getScore(),
                newFeatureGroup);
        // ensure any additional properties are copied
        if (sf.otherDetails != null)
        {
          newSf.otherDetails = new HashMap<String, Object>(sf.otherDetails);
        }
        ffile.parseDescriptionHTML(newSf, false);
        // add any additional links not parsed from description
        if (sf.links != null)
        {
          for (String link : sf.links)
          {
            newSf.addLink(link);
          }
        }
        // amend features only gets one sequence to act on
        sequences.get(0).addSequenceFeature(newSf);

        if (typeChanged)
        {
          findAllFeatures();
        }
      }
    }
    else
    // NEW FEATURES ADDED
    {
      if (reply == JvOptionPane.OK_OPTION && lastFeatureAdded.length() > 0)
      {
        for (int i = 0; i < sequences.size(); i++)
        {
          SequenceFeature sf = features.get(i);
          sf.type = lastFeatureAdded;
          // fix for JAL-1538 - always set feature group here
          sf.featureGroup = lastFeatureGroupAdded;
          sf.description = lastDescriptionAdded;
          sequences.get(i).addSequenceFeature(sf);
          ffile.parseDescriptionHTML(sf, false);
        }

        if (lastFeatureGroupAdded != null)
        {
          setGroupVisibility(lastFeatureGroupAdded, true);
        }
        setColour(lastFeatureAdded, fcol);
        setVisible(lastFeatureAdded);

        findAllFeatures(false);

        alignPanel.paintAlignment(true);

        return true;
      }
      else
      {
        return false;
      }
    }

    alignPanel.paintAlignment(true);

    return true;
  }

  /**
   * update the amend feature button dependent on the given style
   * 
   * @param bigPanel
   * @param col
   * @param col
   */
  protected void updateColourButton(JPanel bigPanel, JLabel colour,
          FeatureColourI col)
  {
    colour.removeAll();
    colour.setIcon(null);
    colour.setToolTipText(null);
    colour.setText("");

    if (col.isSimpleColour())
    {
      colour.setBackground(col.getColour());
    }
    else
    {
      colour.setBackground(bigPanel.getBackground());
      colour.setForeground(Color.black);
      FeatureSettings.renderGraduatedColor(colour, col);
    }
  }

  /**
   * Orders features in render precedence (last in order is last to render, so
   * displayed on top of other features)
   * 
   * @param order
   */
  public void orderFeatures(Comparator<String> order)
  {
    Arrays.sort(renderOrder, order);
  }
}
